/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * Copyright (C) 2003 XQuark Group.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xquery.parser;

import java.util.ArrayList;
import java.util.HashMap;

import org.xquark.schema.SchemaManager;
import org.xquark.xquery.parser.util.Constants;
import org.xquark.xquery.parser.util.DependNode;
import org.xquark.xquery.parser.util.TransformXQueryExpression;
import org.xquark.xquery.typing.QType;

public abstract class XQueryExpression implements Cloneable {
	private static final String RCSRevision = "$Revision: 1.36 $";
	private static final String RCSName = "$Name:  $";

	// for printing only
	protected boolean parenthesis = false;
	protected boolean context = false;
	// for ordering
	protected int[] order = null;
	// for typing
	protected ArrayList parentExpression = null;
	//	protected XQueryExpression parentExpression = null;
	protected QType qtype = null; // org.xquark.xml.typing.QType
	protected ArrayList xtrees = null;

	// for reconstruction
	protected ArrayList skolemIDs = null;
	protected ArrayList dependIDs = null;
	protected boolean root = false;
	protected boolean loop = false;

	//protected Set loopIDs = null;
	protected String loopIDs = null;

	// for referencing
	protected XQueryModule parentModule = null;

	// for decomposer : vector of sources matching the path
	protected ArrayList sourceNames = null;
	protected ArrayList urls = null;

	// to avoid building multiple string representations
	protected String stringValue = null;
	//boolean  modified = false;
	// #############################################################################
	// VISITOR STUFF
	// #############################################################################

	public void accept(ParserVisitor visitor) throws XQueryException {
		visitor.visit(this);
	}

	// ******************************************
	// FOR ACCESS TO PRIVATE/PROTECTED ATTRIBUTES
	// ******************************************

	// parentExpression
	public ArrayList getParentExpression() {
		return parentExpression;
	}
	public void setParentExpression(XQueryExpression expression) {
		if (expression == null)
			return;
		if (parentExpression == null)
			parentExpression = new ArrayList(1);
		else
			parentExpression.clear();
		parentExpression.add(expression);
	}
	public void setParentExpressions(ArrayList parentExpression) {
		this.parentExpression = parentExpression;
	}

	// qtype
	public QType getQType() {
		return qtype;
	}
	public void setQType(QType qtype) {
		this.qtype = qtype;
	}

	// qtype
	public ArrayList getXTrees() {
		return xtrees;
	}
	public void setXTrees(ArrayList xtrees) {
		this.xtrees = xtrees;
	}

	//parentUnit
	public XQueryModule getParentModule() {
		return parentModule;
	}
	public void setParentModule(XQueryModule parentModule) {
		if (parentModule == null)
			return;
		this.parentModule = parentModule;
	}

	// reconstruction
	public ArrayList getSkolemIDs() {
		return skolemIDs;
	}
	public void setSkolemIDs(ArrayList skolemIDs) {
		if (this.skolemIDs == null)
			this.skolemIDs = skolemIDs;
		else if (skolemIDs != null) {
			for (int i = 0; i < skolemIDs.size(); i++) {
				Variable tmpVar = (Variable) skolemIDs.get(i);
				if (!(this.skolemIDs.contains(tmpVar)))
					this.skolemIDs.add(0, tmpVar);
			}
		}
	}

	public ArrayList getDependIDs() {
		return dependIDs;
	}
	public void setDependIDs(ArrayList dependIDs) {
		this.dependIDs = dependIDs;
	}

	public boolean getRoot() {
		return root;
	}

	//    public void setRoot(boolean root,boolean force) {
	//        if (force) this.root = root;
	//        else setRoot(root);
	//    }

	public void setRoot(boolean root) {
		if ((skolemIDs == null || skolemIDs.isEmpty()) && (dependIDs == null || dependIDs.isEmpty()))
			this.root = root;
		if (root && dependIDs != null && !dependIDs.isEmpty()) {
			this.loop = true;
			//dependIDs = null;
		}
	}

	public boolean getLoop() {
		return loop;
	}
	//	public void setRoot(boolean loop) { this.loop = loop; }

	//public Set getLoopIDs() { return loopIDs; }
	//public void setLoopIDs(Set loopIDs) { this.loopIDs = loopIDs; }
	public String getLoopIDs() {
		return loopIDs;
	}
	public void setLoopIDs(String loopIDs) {
		this.loopIDs = loopIDs;
	}

	// parenthesis
	public boolean getParenthesis() {
		return parenthesis;
	}
	public void setParenthesis(boolean parenthesis) {
		this.parenthesis = parenthesis;
	}

	// order
	public int[] getOrder() {
		return order;
	}
	public int getOrder(int index) {
		if (order == null)
			return Constants.NOTHING;
		return order[index];
	}
	public void setOrder(int[] order) {
		this.order = order;
	}
	public void setOrder(int order) {
		if (this.order == null) {
			this.order = new int[2];
			this.order[1] = Constants.NOTHING;
		}
		this.order[0] = order;
	}
	// context
	public boolean getContext() {
		return context;
	}
	public void setContext(boolean context) {
		this.context = context;
	}

	// name if exist
	public String getName() {
		return null;
	}

	// sequenceType if exist
	public void setTypeDeclaration(SequenceType typeDeclaration) {
		return;
	}

	// typeName if exist
	public void setTypeName(QName atomicType) {
		return;
	}

	// expression if exist
	public void setExpression(XQueryExpression expression) throws TypeException, XQueryException {
		return;
	}

	// bindingType if exist
	public void setBindingType(int bindingtype) throws XQueryException {
		return;
	}

	// add   27/03/2003
	public XQueryExpression erasePart(XQueryExpression expr) throws XQueryException {
		if (this == expr)
			return null;
		return this;
	}

	// schemaManager
	public SchemaManager getSchemaManager() {
		if (parentModule == null)
			return null;
		return parentModule.getSchemaManager();
	}

	public boolean equals(XQueryExpression expr) {
		/** @todo : It is ugly, but we need equals()!*/
		if (this == expr) {
			return true;
		}
		String thisString = this.toString();
		String exprString = expr.toString();
		if (thisString.equals(exprString)) {
			return true;
		}
		return false;
	}

	public ArrayList getSteps() {
		return null;
	}

//	public void isModified() {
//		modified = true;
//	}

	public String toString(boolean indent) {
		//if ( /*modified ||*/ stringValue == null) { 
			PrintVisitor pv = new PrintVisitor();
			pv.reset(indent);
			try {
				this.accept(pv);
			} catch (XQueryException qe) {}
			//modified = false;
			stringValue = pv.getString();
//		}
		return stringValue;		
	}
	public String toString() {
		//if ( /*modified ||*/ stringValue == null) { 
			PrintVisitor pv = new PrintVisitor();
			pv.reset(false);
			try {
				this.accept(pv);
			} catch (XQueryException qe) {}
			//modified = false;
			stringValue = pv.getString();
//		}
		return stringValue;
	}

	// #############################################################################
	// CLONE STUFF
	// #############################################################################

	public void addParentExpression(XQueryExpression expression) {
		if (expression == null)
			return;
		if (parentExpression == null)
			parentExpression = new ArrayList(1);
		parentExpression.add(expression);
	}

	public XQueryExpression shallowClone() {
		try {
			return (XQueryExpression) super.clone();
		} catch (CloneNotSupportedException e) {}
		return null;
	}

	public Object clone(boolean renameVars) throws CloneNotSupportedException {
		try {	
			CloneVisitor cloneVisitor = new CloneVisitor(parentModule.getStaticContext(),renameVars);
			this.accept(cloneVisitor);
			return cloneVisitor.getClone();
		} catch (XQueryException xqe) {
			throw new CloneNotSupportedException("Could not clone : " + this);
		}
	}
	
	public Object clone() throws CloneNotSupportedException {
		try {	
			CloneVisitor cloneVisitor = new CloneVisitor(parentModule.getStaticContext(),true);
			this.accept(cloneVisitor);
			return cloneVisitor.getClone();
		} catch (XQueryException xqe) {
			throw new CloneNotSupportedException("Could not clone : " + this);
		}
	}

	public void setSourceNames(ArrayList sources) {
		this.sourceNames = sources;
	}

	public ArrayList getSourceNames() {
		return sourceNames;
	}

	public void setUrls(ArrayList urls) {
		this.urls = urls;
	}

	public ArrayList getUrls() {
		return urls;
	}

	// #############################################################################
	// XPATH STUFF
	// #############################################################################

	public ArrayList getBoundVariables() throws XQueryException {
		GetBoundVariablesVisitor visitor = new GetBoundVariablesVisitor(true,false);
		accept(visitor);
		return visitor.getOnlyVariables();
	}

	// #############################################################################
	// RESTRUCTURE STUFF
	// #############################################################################

	public void substitute(HashMap map, TypeVisitor typeVisitor) throws XQueryException {
		//org.xquark.xml.misc.Debug.print("START substitute : " + map);
		SubstituteVisitor substituteVisitor = new SubstituteVisitor(map, typeVisitor);
		accept(substituteVisitor);
	}

	public void putSkolemIDs(ArrayList ids) throws XQueryException {
		PutSkolemIDsVisitor putSkolemIDsVisitor = new PutSkolemIDsVisitor(ids);
		accept(putSkolemIDsVisitor);
	}

	public TransformXQueryExpression transformRepository() throws ParseException, TypeException, XQueryException {
		return transformRepository(null);
	}

	public TransformXQueryExpression transformRepository(TypeVisitor typeVisitor) throws ParseException, TypeException, XQueryException {

		//boolean debug = org.xquark.xml.misc.Debug.ON;
		//org.xquark.xml.misc.Debug.ON = false;
		//org.xquark.xml.misc.Debug.print(this,"START TRANSFORM");
		//org.xquark.xml.misc.Debug.print(this,"expr:  "+this);

		//org.xquark.xml.misc.Debug.print(this,"START CANONIZE ");
		ArrayList QDB = null;
		XQueryExpression QMEM = null;
		long time = System.currentTimeMillis();
		try {
			CanonizeVisitor canonizeVisitor = new CanonizeVisitor(typeVisitor);
			time = System.currentTimeMillis();
			accept(canonizeVisitor);
			time = System.currentTimeMillis() - time;
			org.xquark.util.misc.Debug.print("-------------------- >>>>>>>>  Canonisation : " + time + " ms");
			QDB = canonizeVisitor.getQDB();
			QMEM = canonizeVisitor.getQMEM();
		} catch (XQueryException e) {
			throw new ParseException("Invalid expression : " + this + " -> " + e.getMessage());
		}
		//org.xquark.xml.misc.Debug.print(this,"STOP  CANONIZE ");
		//org.xquark.xml.misc.Debug.print(this,"QDB:  "+QDB);
		if (QDB == null)
			throw new ParseException("Invalid expression : " + this);
		if (typeVisitor != null)
			for (int si = 0; si < QDB.size(); si++) {
				try {
					((Variable) QDB.get(si)).getExpression().accept(typeVisitor);
				} catch (XQueryException e) {
					throw new TypeException("Could not type expression : " + ((Variable) QDB.get(si)).getExpression() + " -> " + e.getMessage());
				}
				try {
					((Variable) QDB.get(si)).accept(typeVisitor);
				} catch (XQueryException e) {
					throw new ParseException("Invalid expression : " + ((Variable) QDB.get(si)) + " -> " + e.getMessage());
				}
				time = System.currentTimeMillis() - time;
				org.xquark.util.misc.Debug.print("-------------------- >>>>>>>>  Typing canonized : " + time + " ms");
			}

		for (int si = 0; si < QDB.size(); si++) {
			//org.xquark.xml.misc.Debug.print(this,"QDB["+si+"]: "+((XQueryExpression) QDB.get(si)).getQType());
			org.xquark.util.misc.Debug.print(this, "QDB[" + si + "]: " + ((Variable) QDB.get(si)).getExpression());

		}

		//org.xquark.xml.misc.Debug.print(this,"QMEM: "+QMEM);
		//org.xquark.xml.misc.Debug.print(this,"QMEM.getQType(): "+QMEM.getQType());
		//org.xquark.xml.misc.Debug.print(this,"QMEM.getQType(): "+QMEM.getQType());

//		/***********************************/
//		//ArrayList varList = new ArrayList();
//		if (!special)
//			return new TransformXQueryExpression(QDB, QMEM, null, null);
//		/***********************************/

		ArrayList dependenceList = new ArrayList();
		ArrayList returnVariables = null;
		ArrayList idVariables = null;
		ArrayList keyVars = new ArrayList();
		ArrayList rootDepend = new ArrayList();
		try {
			org.xquark.util.misc.Debug.print(this, "START SPECIAL TREATMENT");
			time = System.currentTimeMillis();
			if (QDB != null) {
				for (int l = QDB.size() - 1; l >= 0; l--) {
					ArrayList tmpDependenceList = new ArrayList();
					int minIndex = -1;
					Variable qdbVari = (Variable) QDB.get(l);
					tmpDependenceList.add(qdbVari);
					FLWRExpression qdbFlwri = (FLWRExpression) qdbVari.getExpression();
					ArrayList qbdFlwriVars = qdbFlwri.getVariables();

					for (int m = qbdFlwriVars.size() - 1; m >= 0; m--) {
						Variable qbdFlwriVari = (Variable) qbdFlwriVars.get(m);
						XQueryExpression qbdFlwriVariExpr = qbdFlwriVari.getExpression();
						if (!QDB.contains(qbdFlwriVariExpr))
							continue;

						if (tmpDependenceList.size() < 2)
							tmpDependenceList.add(0, qbdFlwriVariExpr);
						qbdFlwriVars.remove(m);
					}
					if (tmpDependenceList.size() > 1)
						dependenceList.add(0, tmpDependenceList);
				}

				// treat the dependenceList
				HashMap tmpHash = new HashMap(QDB.size());
				for (int l = QDB.size() - 1; l >= 0; l--) {
					Variable qdbVarl = (Variable) QDB.get(l);
					FLWRExpression qdbFlwrl = (FLWRExpression) qdbVarl.getExpression();
					ArrayList qbdFlwrlVars = qdbFlwrl.getVariables();

					boolean foundParent = false;
					for (int m = l - 1; m >= 0; m--) {
						Variable qdbVarm = (Variable) QDB.get(m);
						FLWRExpression qdbFlwrm = (FLWRExpression) qdbVarm.getExpression();
						ArrayList qbdFlwrmVars = qdbFlwrm.getVariables();

						if (qbdFlwrlVars.containsAll(qbdFlwrmVars)) {
							foundParent = true;
							DependNode parentNode = (DependNode) tmpHash.get(qdbFlwrm);
							DependNode sonNode = (DependNode) tmpHash.get(qdbFlwrl);
							if (sonNode == null) {
								sonNode = new DependNode((DependNode) null, null, qdbFlwrl);
								tmpHash.put(qdbFlwrl, sonNode);
							}
							if (parentNode != null)
								parentNode.addSon(0, sonNode);
							else {
								parentNode = new DependNode((DependNode) null, null, qdbFlwrm);
								parentNode.addSon(0, sonNode);
								tmpHash.put(qdbFlwrm, parentNode);
							}
							// add parent Node variables to keyvars
							for (int n = qbdFlwrmVars.size() - 1; n >= 0; n--) {
								Variable parVarn = (Variable) qbdFlwrmVars.get(n);
								if (!keyVars.contains(parVarn))
									keyVars.add(0, parVarn);
							}
							break;
						}
					}
					if (!foundParent && tmpHash.get(qdbFlwrl) != null)
						rootDepend.add(0,tmpHash.get(qdbFlwrl));
				}

				// get varlist
				GetBoundVariablesVisitor visitor = new GetBoundVariablesVisitor(true,true);
				QMEM.accept(visitor);
				ArrayList boundVariables = visitor.getVariables();
				returnVariables = visitor.getOnlyVariables();
				idVariables = visitor.getIDsVariables();

				/*
				for (int si = 0; si < QDB.size(); si++)
				        //org.xquark.xml.misc.Debug.print(this,"QDB["+si+"]: "+((XQueryExpression) QDB.get(si)).getQType());
				        org.xquark.xml.misc.Debug.print(this,"QDB["+si+"]: "+((Variable) QDB.get(si)).getExpression());
				//org.xquark.xml.misc.Debug.print(this,"QMEM = " + QMEM);
				 */
				org.xquark.util.misc.Debug.print(this, "boundVariables = " + boundVariables);

				org.xquark.util.misc.Debug.print(this, "QDB: " + QDB);
				/*DEBUG*/
				for (int si = 0; si < QDB.size(); si++)
					org.xquark.util.misc.Debug.print(this, "QDB[" + si + "]: " + ((Variable) QDB.get(si)).getExpression());
				org.xquark.util.misc.Debug.print(this, "QMEM: " + QMEM);

				ArrayList allVars = new ArrayList();
				for (int l = 0; l < QDB.size(); l++) {
					FLWRExpression tmpExpr = (FLWRExpression) ((Variable) QDB.get(l)).getExpression();
					ArrayList tmpList = tmpExpr.getVariables();
					for (int m = 0; m < tmpList.size(); m++) {
						Variable tmpVar = (Variable) tmpList.get(m);
						if (!(allVars.contains(tmpVar)))
							allVars.add(tmpVar);
						//else keyVars.add(tmpVar);

						//						if (tmpVar.getBindingType() == Constants.FOR_BINDINGTYPE)
						//							if (!(varList.contains(tmpVar)) /* && boundVariables.contains(tmpVar) */
						//								)
						//								varList.add(tmpVar);
					}
					//org.xquark.xml.misc.Debug.print(this,(Variable)QDB.get(l) + " = " + tmpExpr);
				}
				org.xquark.util.misc.Debug.print(this, "boundVariables = " + boundVariables);
				org.xquark.util.misc.Debug.print(this, "keyVars = " + keyVars);
				org.xquark.util.misc.Debug.print(this, "returnVariables = " + returnVariables);
				org.xquark.util.misc.Debug.print(this, "idVariables = " + idVariables);

				org.xquark.util.misc.Debug.print(this, "QDB: " + QDB);
				/*DEBUG*/
				for (int si = 0; si < QDB.size(); si++)
					org.xquark.util.misc.Debug.print(this, "QDB[" + si + "]: " + ((Variable) QDB.get(si)).getExpression());
				org.xquark.util.misc.Debug.print(this, "QMEM: " + QMEM);

				for (int l = 0; l < QDB.size(); l++) {
					Variable vari = (Variable) QDB.get(l);
					FLWRExpression tmpExpr = (FLWRExpression) vari.getExpression();
					XQueryExpressionSequence tmpSeq = (XQueryExpressionSequence) tmpExpr.getReturnClause();
					ArrayList subExprs = (ArrayList) tmpSeq.getSubExpressions().clone();
					for (int m = 0; m < subExprs.size(); m++) {
						Variable tmpVar = (Variable) subExprs.get(m);
						// if (!(boundVariables.contains(tmpVar)) && !keyVars.contains(tmpVar)) {
						if (!(idVariables.contains(tmpVar)) && !returnVariables.contains(tmpVar)) {
							subExprs.remove(m);
							m--;
						}
					}
					if (subExprs.isEmpty())
						tmpExpr.setReturnClause(new XQueryVoid(null));
					else
						tmpSeq.setSubExpressions(subExprs);
				}

				org.xquark.util.misc.Debug.print(this, "QDB: " + QDB);
				/*DEBUG*/
				for (int si = 0; si < QDB.size(); si++)
					org.xquark.util.misc.Debug.print(this, "QDB[" + si + "]: " + ((Variable) QDB.get(si)).getExpression());
				org.xquark.util.misc.Debug.print(this, "QMEM: " + QMEM);

				//AddLoopInfoVisitor loopVisitor = new AddLoopInfoVisitor(varList);
				AddLoopInfoVisitor loopVisitor = new AddLoopInfoVisitor(keyVars);
				QMEM.accept(loopVisitor);

			}

			time = System.currentTimeMillis() - time;
			org.xquark.util.misc.Debug.print("-------------------- >>>>>>>>  Special treatement : " + time + " ms");
			org.xquark.util.misc.Debug.print(this, "STOP  SPECIAL TREATMENT");
			org.xquark.util.misc.Debug.print(this, "STOP  TRANSFORM");
		} catch (XQueryException e) {
			throw new ParseException("Invalid expression : " + this + " -> " + e.getMessage());
		}
		/***********************************/

		///*DEBUG*/

		org.xquark.util.misc.Debug.print(this, "QDB: " + QDB);
		/*DEBUG*/
		for (int si = 0; si < QDB.size(); si++)
			org.xquark.util.misc.Debug.print(this, "QDB[" + si + "]: " + ((Variable) QDB.get(si)).getExpression());
		org.xquark.util.misc.Debug.print(this, "QMEM: " + QMEM);

		//varList.clear();
		ArrayList varList = new ArrayList(1);
		varList.addAll(returnVariables);
		for (int l = 0; l < idVariables.size(); l++) {
			if (!varList.contains(idVariables.get(l)))
				varList.add(idVariables.get(l));
		}

		org.xquark.util.misc.Debug.print(this, "dependenceList = " + dependenceList);
		org.xquark.util.misc.Debug.print(this, "rootDepend = " + rootDepend);
		org.xquark.util.misc.Debug.print(this, "varList = " + varList);
		org.xquark.util.misc.Debug.print(this, "keyVars = " + keyVars);
		org.xquark.util.misc.Debug.print(this, "returnVariables = " + returnVariables);
		org.xquark.util.misc.Debug.print(this, "idVariables = " + idVariables);
		if (varList != null)
			for (int l = 0; l < varList.size(); l++) {
				ArrayList trees = ((Variable) varList.get(l)).getXTrees();
				org.xquark.util.misc.Debug.print(this, ((Variable) varList.get(l)).toString() + " --> trees --> " + trees);
				ArrayList subPathList = ((Variable) varList.get(l)).getSubPathList();
				org.xquark.util.misc.Debug.print(this, ((Variable) varList.get(l)).toString() + " --> subPathList --> " + subPathList);
				if (subPathList != null)
					for (int m = 0; m < subPathList.size(); m++) {
						org.xquark.util.misc.Debug.print(this, ((XQueryExpression) subPathList.get(m)) + " --> xtrees--> " + ((XQueryExpression) subPathList.get(m)).getXTrees());
						org.xquark.util.misc.Debug.print(this, ((XQueryExpression) subPathList.get(m)) + " --> loopIDs --> " + ((XQueryExpression) subPathList.get(m)).getLoopIDs());
					}
			}
		org.xquark.util.misc.Debug.print(this, "QDB: " + QDB);
		/*DEBUG*/
		for (int si = 0; si < QDB.size(); si++)
			org.xquark.util.misc.Debug.print(this, "QDB[" + si + "]: " + ((Variable) QDB.get(si)).getExpression());
		org.xquark.util.misc.Debug.print(this, "QMEM: " + QMEM);

		//org.xquark.xml.misc.Debug.ON = debug;
		//org.xquark.xml.misc.Trace.setTraceMode(true);

		return new TransformXQueryExpression(QDB, QMEM, rootDepend, varList, returnVariables, idVariables);
	}

		
	public TransformXQueryExpression transformExtractor(TypeVisitor typeVisitor) throws ParseException, TypeException {

		//boolean debug = org.xquark.xml.misc.Debug.ON;
		//org.xquark.xml.misc.Debug.ON = false;
		//org.xquark.xml.misc.Debug.print("START TRANSFORM");

		//org.xquark.util.misc.Debug.print("START CANONIZE ");
		ArrayList QDB = null;
		XQueryExpression QMEM = null;
		long time = System.currentTimeMillis();
		try {
			time = System.currentTimeMillis();
			CanonizeVisitor canonizeVisitor = new CanonizeVisitor(typeVisitor);
			accept(canonizeVisitor);
			time = System.currentTimeMillis() - time;
			//org.xquark.xml.misc.Debug.print("-------------------- >>>>>>>>  Canonisation : " + time + " ms");
			//org.xquark.util.misc.Debug.print("STOP  CANONIZE ");
			QDB = canonizeVisitor.getQDB();
			QMEM = canonizeVisitor.getQMEM();
		} catch (XQueryException e) {
			throw new ParseException("Invalid expression : " + this + " -> " + e.getMessage());
		}
		//org.xquark.xml.misc.Debug.print("QDB:  "+QDB);
		if (QDB == null)
			throw new ParseException("Invalid expression : " + this);

		//org.xquark.util.misc.Debug.print("AFTER CANONIZE -> : " );
		/*DEBUG*/
		/*
		for (int si = 0; si < QDB.size(); si++) {
			//org.xquark.xml.misc.Debug.print("QDB["+si+"]: "+((XQueryExpression) QDB.get(si)).getQType());
			org.xquark.util.misc.Debug.print( "QDB[" + si + "]: " + ((Variable) QDB.get(si)).getExpression());
		}
		org.xquark.util.misc.Debug.print("QMEM: " + QMEM);
		//org.xquark.xml.misc.Debug.print("QMEM.getQType(): "+QMEM.getQType());
		//org.xquark.xml.misc.Debug.print("QMEM.getQType(): "+QMEM.getQType());
		*/
		/***********************************/
		ArrayList varList = new ArrayList();
//		if (!special)
//			return new TransformXQueryExpression(QDB, QMEM, null, varList);

		/***********************************/
		ArrayList returnVariables = null;
		ArrayList idVariables = null;
		ArrayList keyVars = new ArrayList();
		ArrayList rootDepend = new ArrayList();
		XQueryExpression remainingWhereExpr = null;
		try {
			//org.xquark.util.misc.Debug.print("START SPECIAL TREATMENT");
			time = System.currentTimeMillis();
			GetBoundVariablesVisitor gbvv = new GetBoundVariablesVisitor(true,true);
			QMEM.accept(gbvv);
//			// add return variables for orderby stuff
//			gbvv.reset(false,true);
//			gbvv.isQDB(true);
//			for (int i=0;i<QDB.size();i++) {
//				((Variable)QDB.get(i)).getExpression().accept(gbvv);
//			}
//			gbvv.reset(true,true);
//			gbvv.isQDB(false);
			// list of all variables appearing in QMEM (eather idvariables or reconstruction variables)
			ArrayList boundVariables = gbvv.getVariables();
			// list of all reconstruction variables appearing in QMEM
			returnVariables = gbvv.getOnlyVariables();
			// list of all id variables appearing in QMEM
			idVariables = gbvv.getIDsVariables();

			/*
			ALGORITHM
			Iterating on QDBs from bottom to top
			1. elimination of for $x in 1 (added by restructuration or canonization) whenever possible
				conditions :
					must be unique variable
					no subpaths exist
				if let binding and more than one QDB
					do erase
					!!! DANGER !!!
					if first QDB erased and remaining where expression is empty
						put root label on QMEM for reconstruction
			2. replacing QDBs in FLWRs
			example
			let $1 := for $x1 in COLX1, $x2 in COLX2 where C1 order by x3, x4 return $x1, $x2
			let $2 := for $X in $1, $y1 in COLY1, $y2 in COLY2 where C2 order by y3, y4 return $X, $y1, $y2
			becomes
			let $1 := for $x in COL1 return $x
			let $2 := for $x1 in COLX1, $x2 in COLX2, $y1 in COLY1, $y2 in COLY2 where C1 & C2 orderby x3, x4, y4 return $x1, $x2, $y1, $y2 			 
			*/
			if (QDB != null) {
				for (int l = QDB.size() - 1; l >= 0; l--) {
					Variable qdbVari = (Variable) QDB.get(l);
					FLWRExpression qdbFlwri = (FLWRExpression) qdbVari.getExpression();
					ArrayList qbdFlwriVars = qdbFlwri.getVariables();
					ArrayList qbdFlwriOrders = qdbFlwri.getOrderBy();

					// some special treatment to avoid having for $x in 1 !!!! ATTENTION !!!
					if (qbdFlwriVars.size() == 1) {
						Variable qbdFlwriVar0 = (Variable) qbdFlwriVars.get(0);
						ArrayList paths = qbdFlwriVar0.getSubPathList();
						if (paths == null || paths.isEmpty()) {
							if (qbdFlwriVar0.getBindingType() == Constants.LET_BINDINGTYPE && QDB.size() > 1) {
								//erasing from QDB
								QDB.remove(l);
								l--;
								// remember where expression if it is the last QBD (it is a silly expression anyway)
								if (QDB.isEmpty())
									remainingWhereExpr = qdbFlwri.getWhereClause();
								// erasing from idvariables
								idVariables.remove(qbdFlwriVar0);
								// erasing from skolemIDs
								EraseSkolemIDVisitor eraseSkolemIDVisitor = new EraseSkolemIDVisitor(qbdFlwriVar0);
								QMEM.accept(eraseSkolemIDVisitor);
								if (l < 0 && remainingWhereExpr == null)
									QMEM.setRoot(true);
								continue;
							} else if (!returnVariables.contains(qbdFlwriVar0) && qbdFlwriVar0.getExpression().getQType().isAtom() && qdbFlwri.getWhereClause() == null) {
								// erasing from other variables
								for (int m = QDB.size() - 1; m > l; m--) {
									ArrayList tmpList = ((FLWRExpression) ((Variable) QDB.get(m)).getExpression()).getVariables();
									tmpList.remove(qbdFlwriVar0);
								}
								//erasing from QDB
								QDB.remove(l);
								l--;
								// erasing from idvariables
								idVariables.remove(qbdFlwriVar0);
								// erasing from skolemIDs
								EraseSkolemIDVisitor eraseSkolemIDVisitor = new EraseSkolemIDVisitor(qbdFlwriVar0);
								QMEM.accept(eraseSkolemIDVisitor);
								if (l < 0)
									QMEM.setRoot(true);
								continue;
							}
						}
					}

					for (int m = qbdFlwriVars.size() - 1; m >= 0; m--) {
						Variable qbdFlwriVari = (Variable) qbdFlwriVars.get(m);
						//org.xquark.xml.misc.Debug.print("qbdFlwriVari = " + qbdFlwriVari);
						XQueryExpression qbdFlwriVariExpr = qbdFlwriVari.getExpression();
						//org.xquark.xml.misc.Debug.print(this,"qbdFlwriVariExpr = " + qbdFlwriVariExpr);
						if (QDB.contains(qbdFlwriVariExpr)) {
							// used[QDB.indexOf(qbdFlwriVariExpr)] = true;
							///*DEBUG*/   for (int z=0;z<QDB.size();z++) //org.xquark.xml.misc.Debug.print(this,"used["+z+"] = " +used[z]);
							//org.xquark.xml.misc.Debug.print(QDB + " CONTAINS " + qbdFlwriVariExpr);
							//org.xquark.xml.misc.Debug.print("qdbFlwri = " + qdbFlwri);
							FLWRExpression tmpFlwr = (FLWRExpression) ((Variable) qbdFlwriVariExpr).getExpression();
							ArrayList tmpList = tmpFlwr.getVariables();
							qbdFlwriVars.remove(m);
							//org.xquark.xml.misc.Debug.print("qdbFlwri = " + qdbFlwri);
							//org.xquark.xml.misc.Debug.print("tmpList = " + tmpList);
							int add = 0;
							for (int n = 0; n < tmpList.size(); n++) {
								Variable tmpVar = (Variable) tmpList.get(n);
								if (!qbdFlwriVars.contains(tmpVar)) {
									qdbFlwri.insertVariableAtPosition(tmpVar, m + add);
									qbdFlwriVars = qdbFlwri.getVariables();
									add++;
								}
							}
							tmpList = tmpFlwr.getOrderBy();
							if (tmpList != null)
								for (int n = 0; n < tmpList.size(); n++) {
									XQueryExpression tmpExpr = (XQueryExpression) tmpList.get(n);
									if (qbdFlwriOrders == null || !qbdFlwriOrders.contains(tmpExpr)) {
										qdbFlwri.addOrderBy(n, tmpExpr);
										qbdFlwriOrders = qdbFlwri.getOrderBy();
									}
								}
							//org.xquark.xml.misc.Debug.print("add = " + add);
							m += add;
							if (add != 0) {
								//((Variable)(Variable)qbdFlwriVars.get(m)).setBindingType(Constants.OUTERFOR_BINDINGTYPE);
								qdbFlwri.addWhereClause(tmpFlwr.getWhereClause());
							}
						}
					}
				}
			}
			// creating variable dependencies
			// ATTENTION : this could have been done much easier (and faster!) in between the previous algorithms
			// TODO
			/*
			temporary HashMap
			QBD -> dependNode
			
			ALGORITHM
			Iterating on QDBs from bottom to top
				Iterating on QDBs from just above to top
				if all variables of above QDB are contained in QDB
					there is a parent relationship
					if son node does not exist create it with no parent
					if parent node does exist add the son node
					else create the parent node, add the son node
					add all parent variables to keyvars list				
			*/
			if (QDB != null && !QDB.isEmpty()) {
				// make dependency map
				HashMap tmpHash = new HashMap(QDB.size());
				for (int l = QDB.size() - 1; l >= 0; l--) {
					Variable qdbVarl = (Variable) QDB.get(l);
					FLWRExpression qdbFlwrl = (FLWRExpression) qdbVarl.getExpression();
					ArrayList qbdFlwrlVars = qdbFlwrl.getVariables();

					boolean foundParent = false;
					for (int m = l - 1; m >= 0; m--) {
						Variable qdbVarm = (Variable) QDB.get(m);
						FLWRExpression qdbFlwrm = (FLWRExpression) qdbVarm.getExpression();
						ArrayList qbdFlwrmVars = qdbFlwrm.getVariables();

						if (qbdFlwrlVars.containsAll(qbdFlwrmVars)) {
							foundParent = true;
							DependNode parentNode = (DependNode) tmpHash.get(qdbFlwrm);
							DependNode sonNode = (DependNode) tmpHash.get(qdbFlwrl);
							if (sonNode == null) {
								sonNode = new DependNode((DependNode) null, null, qdbFlwrl);
								tmpHash.put(qdbFlwrl, sonNode);
							}
							if (parentNode != null)
								parentNode.addSon(0, sonNode);
							else {
								parentNode = new DependNode((DependNode) null, null, qdbFlwrm);
								parentNode.addSon(0, sonNode);
								tmpHash.put(qdbFlwrm, parentNode);
							}
							// add parent Node variables to keyvars
							for (int n = qbdFlwrmVars.size() - 1; n >= 0; n--) {
								Variable parVarn = (Variable) qbdFlwrmVars.get(n);
								if (!keyVars.contains(parVarn))
									keyVars.add(0, parVarn);
							}
							break;
						}
					}
					if (!foundParent && tmpHash.get(qdbFlwrl) != null)
						rootDepend.add(0,tmpHash.get(qdbFlwrl));
				}

				/*
				// DEBUG
				for (int si = 0; si < QDB.size(); si++)
				        //org.xquark.xml.misc.Debug.print("QDB["+si+"]: "+((XQueryExpression) QDB.get(si)).getQType());
				        org.xquark.xml.misc.Debug.print("QDB["+si+"]: "+((Variable) QDB.get(si)).getExpression());
				//org.xquark.xml.misc.Debug.print("QMEM = " + QMEM);
				 */
				//org.xquark.xml.misc.Debug.print(this,"boundVariables = " + boundVariables);

				// TODO check if this code is useful?!!!
				/*
				adds variables to allvars
				adds variables to varlist (only for binding?)
				*/
				ArrayList allVars = new ArrayList();
				for (int l = 0; l < QDB.size(); l++) {
					FLWRExpression tmpExpr = (FLWRExpression) ((Variable) QDB.get(l)).getExpression();
					ArrayList tmpList = tmpExpr.getVariables();
					for (int m = 0; m < tmpList.size(); m++) {
						Variable tmpVar = (Variable) tmpList.get(m);
						if (!(allVars.contains(tmpVar)))
							allVars.add(tmpVar);
						//else keyVars.add(tmpVar);
						if (tmpVar.getBindingType() == Constants.FOR_BINDINGTYPE)
							if (!(varList.contains(tmpVar)))
								varList.add(tmpVar);
					}
					//org.xquark.xml.misc.Debug.print((Variable)QDB.get(l) + " = " + tmpExpr);
				}
				
				// TODO check if this code is useful?!!!
				/*
				erases variables in return clause that are not bound or keys
				if no variables left keep the first
				*/
				for (int l = 0; l < QDB.size(); l++) {
					Variable vari = (Variable) QDB.get(l);
					FLWRExpression tmpExpr = (FLWRExpression) vari.getExpression();
					XQueryExpressionSequence tmpSeq = (XQueryExpressionSequence) tmpExpr.getReturnClause();
					ArrayList subExprs = (ArrayList) tmpSeq.getSubExpressions().clone();
					for (int m = 0; m < subExprs.size(); m++) {
						Variable tmpVar = (Variable) subExprs.get(m);
						if (!(boundVariables.contains(tmpVar)) && !keyVars.contains(tmpVar)) {
							subExprs.remove(m);
							m--;
						}
					}
					if (subExprs.isEmpty()) {
						subExprs.add(tmpSeq.getSubExpressions().get(0));
						tmpSeq.setSubExpressions(subExprs);
					} else
						tmpSeq.setSubExpressions(subExprs);
					//org.xquark.xml.misc.Debug.print((Variable)QDB.get(l) + " = " + tmpExpr);
				}

				SetVariablePathsVisitor varPathVisitor = new SetVariablePathsVisitor();
				for (int l = 0; l < QDB.size(); l++) {
					((Variable)QDB.get(l)).getExpression().accept(varPathVisitor);
				}
				QMEM.accept(varPathVisitor);

			}
			// treat special case if QDB is empty
			// to handle expression calculator like!!!!
			// add fake variable
//			if (QDB != null && QDB.isEmpty()) {
//				Variable var = createVariable(Constants.FOR_BINDINGTYPE, new ValueInteger("1", typeVisitor), typeVisitor);
//				ArrayList vars = new ArrayList(1);
//				vars.add(var);
//				varList.add(var);
//				FLWRExpression tmpFLWR = new FLWRExpression(vars, null, remainingWhereExpr, var, typeVisitor);
//				Variable newVar = createVariable(Constants.LET_BINDINGTYPE, tmpFLWR, typeVisitor);
//				QDB.add(newVar);
//			}
			// special verification for extractor : verify that all variables in varlist have at least one path associated
			// if not add the variable itself (but only for variable bound on Values)
			for (int l = 0; l < varList.size(); l++) {
				Variable var = (Variable) varList.get(l);
				ArrayList pathslist = var.getSubPathList();
				if (pathslist == null || pathslist.isEmpty())
					var.addSubPath(var);
			}

			time = System.currentTimeMillis() - time;
			//org.xquark.xml.misc.Debug.print("-------------------- >>>>>>>>  Special treatement : " + time + " ms");

			//org.xquark.xml.misc.Debug.print("STOP  SPECIAL TREATMENT");
			//org.xquark.xml.misc.Debug.print("STOP  TRANSFORM");
		} catch (XQueryException e) {
			throw new ParseException("Invalid expression : " + this + " -> " + e.getMessage());
		}
		/***********************************/

		///*DEBUG*/
		
		org.xquark.util.misc.Debug.print("rootDepend = " + rootDepend);
		org.xquark.util.misc.Debug.print("varList = " + varList);
		org.xquark.util.misc.Debug.print("keyVars = " + keyVars);
		org.xquark.util.misc.Debug.print("returnVariables = " + returnVariables);
		org.xquark.util.misc.Debug.print("idVariables = " + idVariables);
		if (varList != null)
			for (int l = 0; l < varList.size(); l++) {
				ArrayList trees = ((Variable) varList.get(l)).getXTrees();
				org.xquark.util.misc.Debug.print(((Variable) varList.get(l)) + " --> trees --> " + trees);
				ArrayList subPathList = ((Variable) varList.get(l)).getSubPathList();
				org.xquark.util.misc.Debug.print(((Variable) varList.get(l)) + " --> subPathList --> " + subPathList);
				try {
					org.xquark.util.misc.Debug.print(((Variable) varList.get(l)) + " --> unique subPathList --> " + ((Variable) varList.get(l)).getUniqueSubPathList());
				} catch (XQueryException e) {}
				if (subPathList != null)
					for (int m = 0; m < subPathList.size(); m++) {
						org.xquark.util.misc.Debug.print(((XQueryExpression) subPathList.get(m)) + " --> xtrees--> " + ((XQueryExpression) subPathList.get(m)).getXTrees());
						org.xquark.util.misc.Debug.print(((XQueryExpression) subPathList.get(m)) + " --> loopIDs --> " + ((XQueryExpression) subPathList.get(m)).getLoopIDs());
					}
			}
		
		/*DEBUG*/
		
		for (int si = 0; si < QDB.size(); si++)
			org.xquark.util.misc.Debug.print("QDB[" + si + "]: " + ((Variable) QDB.get(si)).getExpression());
		org.xquark.util.misc.Debug.print("QMEM: " + QMEM);
		
		//org.xquark.xml.misc.Debug.ON = debug;
		return new TransformXQueryExpression(QDB, QMEM, (rootDepend.isEmpty())?null:rootDepend, varList, returnVariables, idVariables);
	}

	private int varIndex = 0;
	private Variable createVariable(int bindingType, XQueryExpression expr, TypeVisitor typeVisitor) throws XQueryException {
		if (typeVisitor != null && expr != null)
			expr.accept(typeVisitor);
		Variable var = new Variable(new QName(null, null, "var_trs" + (++varIndex), expr.getParentModule()), expr.getParentModule());
		var.setBindingType(bindingType);
		var.setExpression(expr);
		return var;
	}

}
